#include <getopt.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <pcap.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "dns.h"
#include "interface.h"
#include "parsing.h"
#include "print.h"
#include "scan.h"
#include "thread.h"

static int scan_host(struct scan *general, uint8_t nb_threads)
{
	if (dns_lookup(general->dest_addr, general->dest_addr)) {
		return -1;
	}

	struct response responses[1024] = {0};
	struct timespec start, end;

	general->responses = responses;
	printf("Scanning...\n");
	clock_gettime(CLOCK_MONOTONIC, &start);
	if (create_threads(general, nb_threads) < 0)
		return -1;
	clock_gettime(CLOCK_MONOTONIC, &end);
	double time_elapsed =
	    (end.tv_sec - start.tv_sec) + (end.tv_nsec - start.tv_nsec) / 1e9;
	print_host_results(general, time_elapsed);
	return 0;
}

int main(int ac, char **av)
{
	if (ac < 2) {
		dprintf(2, "Usage: ft_nmap [Scan Type(s)] [Options] {target "
		           "specification}\n");
		print_usage();
		return 0;
	}

	struct option_lst *options = parse_options(ac, av);
	if (options == NULL) {
		print_usage();
		return 1;
	}

	if (option_isset(options, FL_HELP)) {
		free_options(options);
		print_usage();
		return 0;
	}

	struct scan general;
	if (parsing(&general, options)) {
		free_options(options);
		return 1;
	}

	if (get_interface_name(&general.host) < 0)
		return -1;

	char *nb_threads_str = get_option_arg(options, FL_SPEEDUP);
	uint8_t nb_threads = 1;
	if (nb_threads_str)
		nb_threads = atoi(nb_threads_str);

	char *dest_addr = get_option_arg(options, FL_IP);
	if (dest_addr) {
		general.dest_addr = dest_addr;
		print_config(&general, NULL, nb_threads);
		int rv = scan_host(&general, nb_threads);
		free_options(options);
		return rv;
	}

	const char *hosts_path = get_option_arg(options, FL_FILE);
	if (!hosts_path) {
		dprintf(2, "ft_nmap: address/hostname required\n");
		goto error;
	}

	FILE *hosts_file = fopen(hosts_path, "r");
	if (hosts_file == NULL) {
		dprintf(2, "ft_nmap: unable to open file '%s'\n", hosts_path);
		goto error;
	}

	print_config(&general, hosts_path, nb_threads);
	char line[256];
	while (fgets(line, sizeof(line), hosts_file)) {
		line[strcspn(line, "\n")] = '\0';
		general.dest_addr = line;
		if (scan_host(&general, nb_threads) < 0) {
			fclose(hosts_file);
			goto error;
		}
	}
	fclose(hosts_file);

	free_options(options);
	return EXIT_SUCCESS;
error:
	free_options(options);
	return EXIT_FAILURE;
}
